{-# LANGUAGE FlexibleInstances,
             MultiParamTypeClasses,
             TypeOperators,
             UndecidableInstances,
             ScopedTypeVariables,
             FlexibleContexts
 #-}

module AOP.Internal.AOT (
 AOT,
 runAOT,
 wappt,
) where

import AOP.Internal.Typeable1Monad
import AOP.Internal.Aspect
import AOP.Internal.Pointcut
import AOP.Internal.Advice
import AOP.Internal.Joinpoint
import AOP.Internal.AOPMonad
import AOP.Internal.OpenApp
import AOP.Internal.Function
import AOP.Internal.AOTType


-- | Runs an AOT computation to obtain a computation in the underlying monad
runAOT :: Monad m => AOT m a -> m a
runAOT c = do
             (a, _) <- run c [];
             return a

-- | Monadic weaver
weavet :: (Monad m) => (a -> AOT m b) -> AspectEnv (AOT m) -> Jp a (AOT m b) -> m (a ->  AOT m b)
weavet f [] _ = return f
weavet f aenv@(asp:asps) jp = 
 case asp of EAspect (Aspect pc adv _) -> do (match, _) <- run (runPC pc jp) aenv
                                             weavet (if match  
                                                         then (apply_adv adv f)
                                                         else f)
                                                     asps jp

-- | Implementation of woven application for AOT, used in the overloading of #.
-- FunctionTag argument is used for function identity:
--     Wrapped functions are equal when their tags are equal
--     Regular functions share the same tag, and are compared using StableNames.
wappt :: (Typeable1Monad m, PolyTypeable (a -> AOT m b)) => (a -> AOT m b) -> FunctionTag -> a -> AOT m b
wappt f t a = AOT $ \aenv -> do woven_f <- weavet f aenv (newjp f t a)
                                run (woven_f a) aenv

-- | Every regular functions is tagged with the same default tag.
instance Typeable1Monad m => OpenApp (->) (AOT m) where
         f # a = wappt f defaultFunctionTag a

-- | Function is a wrapper to add a notion of identity based on tags.
instance Typeable1Monad m => OpenApp Function (AOT m) where
         (Function f t) # a = wappt f t a

-- | Monadic bind and return operations for a transformed monad.
-- The aspect environment is wrapped or passed around as in a state monad transformer.
instance Monad m => Monad (AOT m) where
    return a = AOT $ \aenv -> return (a, aenv)
    k >>= f  = AOT $ \aenv -> do (arg, aenv') <- run k aenv
                                 run (f arg) aenv'

-- | Typeable instance so types of computations in AOT can be compared (like in pcCall and pcType)
instance Typeable1Monad m => Typeable1 (AOT m) where
         typeOf1 _ = mkTyConApp (mkTyCon3 "AOT" "AOT" "AOT") [typeOf1 (undefined :: m ())]

-- | The semantics of aspect deployment are defined in the
-- | MonadDeploy typeclass. AOT assumes it is on top of an MonadDeploy
-- | instance, and uses that functions for aspect deployment.
instance (Typeable1Monad m, MonadDeploy m) => AOPMonad (AOT m) where
      deploy asp   = AOT $ \aenv -> do aenv' <- deployInEnv asp aenv
                                       return ((), aenv')
      undeploy asp = AOT $ \aenv -> do aenv' <- undeployInEnv asp aenv
                                       return ((), deleteAsp (EAspect asp) aenv)


-- Interaction with standard monad transformers

instance MonadTrans AOT where
         lift ma = AOT $ \ctx -> do
                           a <- ma
                           return (a, ctx)

instance (MonadIO m) => MonadIO (AOT m) where
         liftIO = lift . liftIO

instance (MonadState s m) => MonadState s (AOT m) where
         get = lift get
         put = lift . put

instance (MonadError s m) => MonadError s (AOT m) where
         throwError = lift . throwError
         -- ma `catchError` h = AOT $ \ctx -> do
         --        (a,ctx') <- run ma ctx
         --        case a of
         --             (Left  l) -> run (h l) ctx'
         --             (Right r) -> run ma ctx' -- return (Right r, ctx')

instance (MonadWriter w m) => MonadWriter w (AOT m) where
    tell     = lift . tell
    listen m = AOT $ \ctx -> do
                              ((a, ctx'), w) <- listen (run m ctx)
                              return ((a, w), ctx')

    pass m = AOT $ \ctx -> pass $ do
                                   ((a, f), ctx') <- run m ctx
                                   return ((a, ctx'), f)

instance MonadCont m => MonadCont (AOT m) where
    callCC f = AOT $ \aenv ->
        callCC $ \c ->
        run (f (\a -> AOT $ \aenv' -> c (a, aenv'))) aenv

instance (MonadReader r m) => MonadReader r (AOT m) where
    ask       = lift ask
    local f m = AOT $ \s -> local f (run m s)



