{-# LANGUAGE FlexibleContexts,
             ExistentialQuantification
 #-}

module AOP.Internal.Advice (
 Advice,
 apply_adv,
 replace,
 augment,
 narrow, 
 before,
 after,
 Narrow
) where

import Unsafe.Coerce

type Advice a = (a -> a)

{- The use of unsafeCoerce below is safe because any function
   that a pc matches is compatible by construction with the advice.
   See Section 4 of the paper.
-}

-- | Coerces t2 to be compatible with the advice. It passes t1 as a the proceed argument of the advice.
-- This coercion is safe, as described in Section 4 of the paper.
apply_adv :: Advice t1 -> t2 -> t2
apply_adv adv f = (unsafeCoerce adv) f


-- Enforcement of control flow properties, adapted from EffectiveAdvice

-- Replacement: no calls to proceed. Computation is replaced by radv.
type Replace s = s
replace :: Replace s -> Advice s
replace radv = \proceed -> radv

-- Augmentation: proceed is called only once. Behavior can be specified before and after
-- the implicit call to proceed. The second part has access to the return value,
-- and to the value returned by the first part.
type Augment a b c m = (a -> m c, a -> b -> c -> m ())
augment :: Monad m => Augment a b c m -> Advice (a -> m b)
augment (before, after) proceed arg = do c <- before arg
                                         b <- proceed arg
                                         after arg b c
                                         return b

-- before/after advice can be expressed as particular cases of augmentation
before before = augment (\a -> before a >> return (), \a b c -> return ())
after  after  = augment (\_ -> return (), \a b c -> after a b)


-- Narrowing: proceed is called at most once. A runtime choice can be
-- made between replacement or augmentation advice. Typical case: memoization
type Narrow a b c m = (a -> m Bool, (a -> m c, a -> b -> c -> m ()), Replace (a -> m b))
narrow :: Monad m => Narrow a b c m -> Advice (a -> m b)
narrow (p, aug, rep) proceed arg = 
       do test <- p arg
          if test
             then (augment aug proceed arg)
             else (replace rep proceed arg)


