{-# LANGUAGE MultiParamTypeClasses,
             TypeFamilies,
             FunctionalDependencies,
             UndecidableInstances,
             FlexibleInstances,
             OverlappingInstances,
             IncoherentInstances
 #-}

module AOP.Internal.LessGen (LessGen, LeastGen, Analyze, LeastGen', SubstEmpty) where

import AOP.Internal.PolyTypeable

-- Check if two analysed types can be matchedd.
-- The tricky part is to construct a substitution using type classes.

-- | If LeastGen a b c holds then there exists a substitution s and type c such that (s c = (a, b)).
class LeastGen a b c

-- LeastGen is defined in terms of LeastGen'. We discard the constructed substitution sout.
instance (Analyze a _a, Analyze b _b, LeastGen' _a _b c SubstEmpty sout) => 
         LeastGen a b c

-- | If LessGen a b, then b is less general than a. That is, there exists substitution s such that (s b = a).
class LessGen a b

-- | It unifies a with b after checking the substitution s.
-- | We use this to delay unification in the following instance declaration for LessGen.
-- | If we just use typical unification, this will be done *before* constructing/checking the substitution.
-- | The idea is that if the substitution exists, then the unification is one-way and is performed.
class Unifies s a b
instance a ~ b => Unifies SubstEmpty a b
instance a ~ b => Unifies (SubstCons x y s) a b

 -- LessGen is also defined in terms of LeastGen'
instance (Analyze a _a, Analyze b _b, LeastGen' _a _b b SubstEmpty sout, Unifies sout a b) 
         => LessGen a b

-- LeastGen' extends the substitution s0 to sout 
-- such that (sout b = a).
class LeastGen' a b c sin sout | sin a b c -> sout

-- Constant types we check if they are both the same
instance LeastGen' (TCon0 a) (TCon0 a) a s0 s0 -- (a ~ d) .... d

-- for n-ary constructors, the substitution is extended for each argument
-- and chained from left to right 
instance (LeastGen' a1 b1 c1 s0 s1,
          d c1 ~ c)
       => LeastGen' (TCon1 d a1) (TCon1 d b1) c s0 s1

instance (LeastGen' a1 b1 c1 s0 s1,
          LeastGen' a2 b2 c2 s1 s2,
          d c1 c2 ~ c)
       => LeastGen' (TCon2 d a1 a2) (TCon2 d b1 b2) c s0 s2

instance (LeastGen' a1 b1 c1 s0 s1,
          LeastGen' a2 b2 c2 s1 s2,
          LeastGen' a3 b3 c3 s2 s3,
          d c1 c2 c3 ~ c)
       => LeastGen' (TCon3 d a1 a2 a3) (TCon3 d b1 b2 b3) c s0 s3

instance (LeastGen' a1 b1 c1 s0 s1,
          LeastGen' a2 b2 c2 s1 s2,
          LeastGen' a3 b3 c3 s2 s3,
          LeastGen' a4 b4 c4 s3 s4,
          d c1 c2 c3 c4 ~ c)
       => LeastGen' (TCon4 d a1 a2 a3 a4) (TCon4 d b1 b2 b3 b4) c s0 s4

instance (LeastGen' a1 b1 c1 s0 s1,
          LeastGen' a2 b2 c2 s1 s2,
          LeastGen' a3 b3 c3 s2 s3,
          LeastGen' a4 b4 c4 s3 s4,
          LeastGen' a5 b5 c5 s4 s5,
          d c1 c2 c3 c4 c5 ~ c)
       => LeastGen' (TCon5 d a1 a2 a3 a4 a5) (TCon5 d b1 b2 b3 b4 b5) c s0 s5

-- default case when a and b doesn't share the same constructor
instance (MapsTo sin (a, b) c',
          VarCase c' (a, b) c sin sout,
          Analyze c (TVar c))
       => LeastGen' a b c sin sout

-- extends the substitution if required
class VarCase v ab c sin sout | v ab sin -> sout c
instance VarCase None ab c sin (SubstCons ab c sin)
instance VarCase (Some c) ab c sin sin -- check that s(b) = a

----------------------------------

-- | Encoding of substitutions as partial maps 
class Substitution s where 
     data SubstEmpty
     data SubstCons x sx s

class MaybeValue a where
     data None
     data Some a

-- | Examines substitution s and binds sx the variable that maps to x in s, or None.
class MapsTo s x sx | s x -> sx

instance MapsTo SubstEmpty x None
instance MapsTo (SubstCons x sx s) x (Some sx) 
instance MapsTo s x sx => MapsTo (SubstCons x' sx' s) x sx
