{-# LANGUAGE FlexibleInstances,
             MultiParamTypeClasses,
             TypeOperators,
             UndecidableInstances,
             ScopedTypeVariables,
             FlexibleContexts,
             KindSignatures
 #-}

module AOP.Internal.NIAOT (
 NIAOT,
 runNIAOT,
 niLift
) where

import AOP.Internal.Typeable1Monad
import AOP.Internal.Aspect
import AOP.Internal.Pointcut
import AOP.Internal.Advice
import AOP.Internal.Joinpoint
import AOP.Internal.AOPMonad
import AOP.Internal.OpenApp
import AOP.Internal.Function

{- | Non-Interference AOT transformer.
The NIAOT is like the AOT transformer, but it enforces a splitting on the monadic stack.
Base code and aspect code have access to a different part of the stack.
-}


-- | The splitting is done by assuming the stack is constructed by a transformer t on top of a monad m.
newtype NIAOT (t :: (* -> *) -> * -> *) m a = NIAOT {runNI :: AspectEnv (NIAOT t m) -> t m (a, AspectEnv (NIAOT t m))}

-- | Gets the underlying computation
runNIAOT :: (Monad (t m)) => NIAOT t m a -> t m a
runNIAOT c = do
             (a, _) <- runNI c [];
             return a

-- | Monadic weaver
weavet :: (Monad (t m)) => (a -> NIAOT t m b) -> AspectEnv (NIAOT t m) -> 
                           Jp a (NIAOT t m b) -> t m (a ->  NIAOT t m b)
weavet f [] _ = return f
weavet f aenv@(asp:asps) jp = 
    case asp of EAspect (Aspect pc adv _) -> 
                            do                               
                               (match, _) <- runNI (runPC pc jp) aenv
                               weavet (if match
                                        then (apply_adv adv f)
                                        else f)
                                       asps jp

-- | Implementation of woven application for NIAOT, used in the overloading of #.
-- | Function tags are incorporated like in AOT.
wappt :: (Typeable1Monad (t m), PolyTypeable (a -> NIAOT t m b)) => 
         (a -> NIAOT t m b) -> FunctionTag -> a -> NIAOT t m b
wappt f t a = NIAOT $ \ctx -> do woven_f <- weavet f ctx (newjp f t a)
                                 runNI (woven_f a) ctx

-- | Monadic bind and return functions of the transformed monad.
instance (Monad (t m)) => Monad (NIAOT t m) where
    return a = NIAOT $ \aenv -> return (a, aenv)
    k >>= f  = NIAOT $ \aenv ->
                           do
                              (arg, aenv') <- runNI k aenv
                              runNI (f arg) aenv'

-- | Open application of regular functions.
instance Typeable1Monad (t m) => OpenApp (->) (NIAOT t m) where
         f # a = wappt f defaultFunctionTag a

-- | Open application of tagged functions.
instance Typeable1Monad  (t m) => OpenApp Function (NIAOT t m) where
         (Function f t) # a = wappt f t a

-- | Lift a computation from the underlying monad into NIAOT.
niLift ma = NIAOT $ \ctx -> do
                           a <- ma
                           return (a, ctx)

-- | NIAOT performs top-level deployment of aspects
instance (Typeable1Monad (t m)) => AOPMonad (NIAOT t m) where
      deploy asp   = NIAOT $ \aenv -> return ((), EAspect asp : aenv)
      undeploy asp = NIAOT $ \aenv -> return ((), deleteAsp (EAspect asp) aenv)

-- | Support for PolyTypeable
instance (Typeable1Monad (t m)) => Typeable1 (NIAOT t m) where
         typeOf1 _ = mkTyConApp (mkTyCon3 "NIAOT" "NIAOT" "NIAOT") 
                     [typeOf1 (undefined :: t m ())]


-- Interaction with standard monad transformers

instance (MonadIO (t m)) => MonadIO (NIAOT t m) where
         liftIO = niLift . liftIO

instance (MonadState s (t m)) => MonadState s (NIAOT t m) where
         get = niLift get
         put = niLift . put

instance (MonadError s (t m)) => MonadError s (NIAOT t m) where
         throwError = niLift . throwError
         -- catchError?

instance (MonadWriter w (t m)) => MonadWriter w (NIAOT t m) where
    tell     = niLift . tell
    listen m = NIAOT $ \ctx -> do
                              ((a, ctx'), w) <- listen (runNI m ctx)
                              return ((a, w), ctx')
    pass m = NIAOT $ \ctx -> pass $ do
                                   ((a, f), ctx') <- runNI m ctx
                                   return ((a, ctx'), f)

instance (MonadCont (t m)) => MonadCont (NIAOT t m) where
    callCC f = NIAOT $ \aenv ->
        callCC $ \c ->
        runNI (f (\a -> NIAOT $ \aenv' -> c (a, aenv'))) aenv

instance (Monad m, MonadReader r (t m)) => MonadReader r (NIAOT t m) where
    ask       = niLift ask
    local f m = NIAOT $ \s -> local f (runNI m s)
