{-# LANGUAGE FlexibleContexts,
             ConstraintKinds,
             TypeFamilies,
             MultiParamTypeClasses,
             FlexibleInstances
 #-}


module AOP.Internal.PointcutDef (
 pcCall,
 pcType,
 pcAnd,
 pcOr,
 pcNot,
 pcSeq
) where

import AOP.Internal.Typeable1Monad
import AOP.Internal.Pointcut
import AOP.Internal.Joinpoint
import AOP.Internal.LessGen
import AOP.Internal.Function
import GHC.Prim(Constraint)

{- |
Built-in pointcuts pcCall and pcType, and pointcut combinators pcAnd, pcOr and pcNot.
Using typeclasses, pointcuts are open for new definitions of functions, like the Function wrapper.
We also define pcSeq, that matches a sequence of two join points.
-}

_pcCall f t tag = return (\jp -> return (compareFun f tag jp && compareType t jp))

-- | The class of call pointcuts.
class PcCall f where
      pcCall :: (Monad m, PolyTypeable (f a b)) => f a b -> (PC m (a -> b))

-- | Call pointcut for regular functions, which have a shared default function tag. We use StableNames for function equality.
instance PcCall ((->)) where 
         pcCall f = let t = polyTypeOf f in PC (_pcCall f t defaultFunctionTag)

-- | Call pointcut for our encoding of function identity. Functions that are equal have the same tag.
instance PcCall Function where
         pcCall (Function f tag) = let t = polyTypeOf f in PC (_pcCall f t tag)

_pcType t = (return (\jp -> return (compareType t jp)))

-- | Class for type pointcuts.
class PcType f where
      pcType :: (PolyTypeable (f a b), Monad m) => f a b -> (PC m (a -> b))

-- | Type pointcut for regular functions.
instance PcType ((->)) where
         pcType f = let t = polyTypeOf f in PC (_pcType t)

-- | Type pointcut for our encoding of function identity. The function tag has no influence.
instance PcType Function where
         pcType (Function f _) = let t = polyTypeOf f in PC (_pcType t)

pcCombinator op mpc1 mpc2 =  (do  pc1 <- mpc1
                                  pc2 <- mpc2
                                  return (\jp ->
                                           do
                                            res1 <- pc1 jp
                                            res2 <- pc2 jp
                                            return (res1 `op` res2)))

pcCombinator' op mpc1 =  (do  pc1 <- mpc1
                              return (\jp ->
                                   do
                                      res1 <- pc1 jp
                                      return (op res1)))

-- | And pointcut combinator, overloaded to support PC and RequirePC
class Monad m => PCAnd m t1 t2 pct where
  type PCAndCtx m t1 t2 pct :: Constraint
  pcAnd :: PCAndCtx m t1 t2 pct => PC m t1 -> pct m t2 -> PC m t1

-- | When combining two PC pointcuts, the matched types t1 and t2 must be the same
-- This is expressed in the constraint t1 ~ t2

-- (Note: if we use only t instead of t1 t2 and t1 ~ t2, we get a compile error in AOTTests.hs)
instance Monad m => PCAnd m t1 t2 PC where
         type PCAndCtx m t1 t2 PC = (t1 ~ t2)
         pcAnd (PC mpc1) (PC mpc2) = PC (pcCombinator (&&) mpc1 mpc2)

-- | When combining a PC with a RequirePC we constraint t1 to be LessGen than t2
instance Monad m => PCAnd m t1 t2 RequirePC where
         type PCAndCtx m t1 t2 RequirePC = (LessGen t1 t2)
         pcAnd (PC mpc1) (RequirePC mpc2) = PC (pcCombinator (&&) mpc1 mpc2)

-- | Or pointcut combinator.
pcOr :: (Monad m, LeastGen t1 t2 tgen) => PC m t1 -> PC m t2 -> PC m tgen
pcOr (PC mpc1) (PC mpc2) = PC (pcCombinator (||) mpc1 mpc2)

-- | Not pointcut combinator.
pcNot :: (Monad m) => PC m t1 -> PC m t2 
pcNot (PC mpc) =  PC (pcCombinator' not mpc)

-- | Matches the sequence of two pcs
pcSeq :: (MonadState Bool m) => PC m t1 -> PC m t2 -> PC m t2
pcSeq (PC mpc1) (PC mpc2) = 
       PC (do  pc1 <- mpc1
               pc2 <- mpc2
               return (\ jp -> do b <- get
                                  if b 
                                    then pc2 jp
                                    else do pc1_ <- pc1 jp
                                            put pc1_
                                            return False))